[llvm] [LoopInterchange] Fix the vectorizable check for a loop (PR #133667)
Ryotaro Kasuga via llvm-commits
llvm-commits at lists.llvm.org
Wed Apr 2 05:13:49 PDT 2025
https://github.com/kasuga-fj updated https://github.com/llvm/llvm-project/pull/133667
>From b1f0744edee4762623081d1d87e7915108ec49b0 Mon Sep 17 00:00:00 2001
From: Ryotaro Kasuga <kasuga.ryotaro at fujitsu.com>
Date: Thu, 27 Mar 2025 09:52:16 +0000
Subject: [PATCH] [LoopInterchange] Fix the vectorizable check for a loop
In the profitability check for vectorization, the dependency matrix was
not handled correctly. This can result to make a wrong decision: It may
say "this loop can be vectorized" when in fact it cannot. The root cause
of this is that the check process early returns when it finds '=' or 'I'
in the dependency matrix. To make sure that we can actually vectorize
the loop, we need to check all the rows of the matrix. This patch fixes
the process of checking whether we can vectorize the loop or not. Now it
won't make a wrong decision for a loop that cannot be vectorized.
Related: #131130
---
.../lib/Transforms/Scalar/LoopInterchange.cpp | 44 ++++++++++++-------
.../profitability-vectorization-heuristic.ll | 9 ++--
2 files changed, 30 insertions(+), 23 deletions(-)
diff --git a/llvm/test/Transforms/LoopInterchange/profitability-vectorization-heuristic.ll b/llvm/test/Transforms/LoopInterchange/profitability-vectorization-heuristic.ll
index efd2a9c09e7cf..0f5aee582373d 100644
--- a/llvm/test/Transforms/LoopInterchange/profitability-vectorization-heuristic.ll
+++ b/llvm/test/Transforms/LoopInterchange/profitability-vectorization-heuristic.ll
@@ -15,16 +15,13 @@
; }
; }
;
-; FIXME: These loops are not exchanged at this time due to the problem in
-; profitability heuristic calculation for vectorization.
-; CHECK: --- !Missed
+; CHECK: --- !Passed
; CHECK-NEXT: Pass: loop-interchange
-; CHECK-NEXT: Name: InterchangeNotProfitable
+; CHECK-NEXT: Name: Interchanged
; CHECK-NEXT: Function: interchange_necessary_for_vectorization
; CHECK-NEXT: Args:
-; CHECK-NEXT: - String: Interchanging loops is not considered to improve cache locality nor vectorization.
-; CHECK-NEXT: ...
+; CHECK-NEXT: - String: Loop interchanged with enclosing loop.
define void @interchange_necessary_for_vectorization() {
entry:
br label %for.i.header
More information about the llvm-commits
mailing list
diff --git a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
index e777f950a7c5a..1dccba4cfa7b8 100644
--- a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
+++ b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
@@ -1197,25 +1197,35 @@ LoopInterchangeProfitability::isProfitablePerInstrOrderCost() {
return std::nullopt;
}
+/// Return true if we can vectorize the loop specified by \p LoopId.
+static bool canVectorize(const CharMatrix &DepMatrix, unsigned LoopId) {
+ for (unsigned I = 0; I != DepMatrix.size(); I++) {
+ char Dir = DepMatrix[I][LoopId];
+ if (Dir != 'I' && Dir != '=')
+ return false;
+ }
+ return true;
+}
+
std::optional<bool> LoopInterchangeProfitability::isProfitableForVectorization(
unsigned InnerLoopId, unsigned OuterLoopId, CharMatrix &DepMatrix) {
- for (auto &Row : DepMatrix) {
- // If the inner loop is loop independent or doesn't carry any dependency
- // it is not profitable to move this to outer position, since we are
- // likely able to do inner loop vectorization already.
- if (Row[InnerLoopId] == 'I' || Row[InnerLoopId] == '=')
- return std::optional<bool>(false);
-
- // If the outer loop is not loop independent it is not profitable to move
- // this to inner position, since doing so would not enable inner loop
- // parallelism.
- if (Row[OuterLoopId] != 'I' && Row[OuterLoopId] != '=')
- return std::optional<bool>(false);
- }
- // If inner loop has dependence and outer loop is loop independent then it
- // is/ profitable to interchange to enable inner loop parallelism.
- // If there are no dependences, interchanging will not improve anything.
- return std::optional<bool>(!DepMatrix.empty());
+ // If the outer loop is not loop independent it is not profitable to move
+ // this to inner position, since doing so would not enable inner loop
+ // parallelism.
+ if (!canVectorize(DepMatrix, OuterLoopId))
+ return false;
+
+ // If inner loop has dependence and outer loop is loop independent then it is
+ // profitable to interchange to enable inner loop parallelism.
+ if (!canVectorize(DepMatrix, InnerLoopId))
+ return true;
+
+ // If both the inner and the outer loop can be vectorized, it is necessary to
+ // check the cost of each vectorized loop for profitability decision. At this
+ // time we do not have a cost model to estimate them, so return nullopt.
+ // TODO: Estimate the cost of vectorized loop when both the outer and the
+ // inner loop can be vectorized.
+ return std::nullopt;
}
bool LoopInterchangeProfitability::isProfitable(